function [CDF, ICDF] = cdfByGPD(x, pLower, pUpper, cdfFunction)
%CDFBYGPD CDF and inverse CDF with generalized Pareto tails.
%   [F,Q] = CDFBYGPD(X, PL, PU) estimates the cumulative distribution function F
%   and its inverse Q with scalar lower and upper Pareto tail probabilities PL 
%   and PU. X is a real-valued vector of data values whose extreme observations 
%   are fit to generalized Pareto distributions (GPD). F is a function handle to
%   the composite CDF with Pareto tails. PL and PU identify the lower and upper
%   tail cumulative probabilities such that 100*PL and 100*(1 - PU) percent of 
%   the observations in X are, respectively, reserved and fit to a GPD by maximum 
%   likelihood. If PL = 0, then no lower Pareto tail is fit. If PU = 1, then no 
%   upper Pareto tail is fit. 
%
%   [F,Q] = CDFBYGPD(X, PL, PU, CDFFUN) uses the function CDFFUN to estimate the 
%   CDF of X between the lower and upper tail probabilities. If specified, 
%   CDFFUN must accept the input data vector X, and return the fitted cumulative
%   probabilities and the data values at which the fitted probabilities are 
%   evaluated. That is, CDFFUN must be of the form [P,XI] = CDFFUN(..., X, ...) 
%   where P is a vector of fitted cumulative probabilities evaluated at the 
%   values in vector XI (ellipses indicate any number of leading and trailing 
%   inputs). Note that the evaluation points in XI do not necessarily need to be
%   the same as the input data in X. If CDFFUN is unspecified, the default is 
%   the empirical cumulative distribution function smoothed by linear 
%   interpolation.
%
%   Notice that the outputs F and Q are not the fitted CDF and inverse CDF, but
%   rather function handles that access the fits of the input vector X as a 
%   reference. This reference fit, although originally based on X, may then be 
%   invoked with an arbitrary data vector Z, and the CDF and inverse CDF of Z 
%   are simply F(Z) and Q(P), respectively. In particular, Z = Q(F(Z)). Note 
%   that a linear interpolation is applied to observations of that fall between 
%   the lower and upper Pareto tails. Since the fit is performed only once, this 
%   function is well-suited to copula and Monte Carlo simulation methods.
%
%   See also ECDF, GPCDF, GPFIT.

%   Copyright 1993-2005 The MathWorks, Inc.
%   $Revision$   $Date$

if (nargin < 3) || (nargin > 4)
    error('stats:cdfByGPD:InvalidNumberOfInputs', ...
          'Specify 3 or 4 input arguments.');
end

if ~strcmpi(class(x), 'double') || ~isreal(x) || (length(x) ~= numel(x))
   error('stats:cdfByGPD:VectorRequired', ...
         'Input data X must be a real-valued vector.');
end

if (numel(pLower) == 0) || (pLower < 0) || (pLower > 1)
   error('stats:cdfByGPD:InvalidPL', ...
         'PL tail probability must be a scalar on [0,1].');
end

if (numel(pUpper) == 0) || (pUpper < 0) || (pUpper > 1)
   error('stats:cdfByGPD:InvalidPU', ...
         'PU tail probability must be a scalar on [0,1].');
end

if pLower > pUpper
   error('stats:cdfByGPD:InconsistentProbabilities', ...
         'Tail probabilities must be such that 0 <= PL <= PU <= 1.');
end

if (nargin < 4) || isempty(cdfFunction)
   cdfFunction = @smoothECDF;
else
   if ~strcmpi(lower(class(cdfFunction)), 'function_handle')
      error('stats:cdfByGPD:NonFunctionHandle', ...
            'CDFFUN must be a function handle.');
   end
end

x = x(:);

%
% Fit a generalized Pareto distribution (GPD) to the tails of the CDF.
%

isLowerTailEmpty = false;      % Assume tails will have some data.
isUpperTailEmpty = false;

if pLower > 0                  % Lower tail estimation via maximum likelihood.

   QL = quantile(x, pLower);   % Lower tail threshold.
   y  = QL - x(x < QL);        % Exceedances below lower threshold (all positive).

   if isempty(y)
      isLowerTailEmpty = true; % Lower tail is empty.
   else
      parameters = gpfit(y);
      lowerTail  = parameters(1); 
      lowerSigma = parameters(2);
   end

end

if pUpper < 1                  % Upper tail estimation via maximum likelihood.

   QU = quantile(x, pUpper);   % Upper tail threshold.
   y  = x(x > QU) - QU;        % Exceedances above upper threshold.

   if isempty(y)
      isUpperTailEmpty = true; % Upper tail is empty.
   else
      parameters = gpfit(y);
      upperTail  = parameters(1); 
      upperSigma = parameters(2);
   end

end

%
% Estimate the empirical CDF function of the input data. This CDF estimation 
% function may, for example, smooth the CDF curve, or simply return the 
% classical "staircase" sample CDF. All that is required is that it return a 
% vector of values of the empirical CDF evaluated at X, F(X).
%

try
  if nargin(cdfFunction) == 3
%
%    Estimate the interior of the CDF using ONLY the observations of the input 
%    data set that fall within the interior (i.e., between the Pareto tails). 
%    In this approach, the interior of the CDF is inserted between the Pareto 
%    tails.
%
%    Constructing the CDF/ICDF in this fashion is designed to minimize any
%    discontinuities at the boundaries between the Pareto tails and smoothed 
%    interior.
%
     [F,X] = cdfFunction(x, QL, QU);
     F     = pLower +  F*(pUpper - pLower);

  else
%
%    Estimate the interior of the CDF using ALL the observations of the input 
%    data set. In this approach, the tails of the CDF are discarded and 
%    over-written with the Pareto tails.
%
%    Constructing the CDF/ICDF in this fashion typically introduces a
%    discontinuity at the boundaries between the Pareto tails and smoothed 
%    interior.
%
     [F,X] = cdfFunction(x);

  end

catch
  error('stats:cdfByGPD:InvalidCDFFUN', 'Error calling function CDFFUN.');
end

X     = X(:);              % Guarantee column vectors.
F     = F(:);
[X,I] = unique(X);         % It should be sorted, but is UNIQUE necessary?
F     = F(I);
xMin  = min(X);
xMax  = max(X);

%
% Now return the actual CDF and inverse CDF functions.
%

CDF  = @getCDF;
ICDF = @getICDF;

%
%  Identify the "out-of-sample" inverse CDF evaluation function.
%
   function xData = getICDF(P)
%  GETICDF Evaluate and return quantiles of the inverse CDF.
%     X = GETICDF(P) evaluates the inverse CDF, returning a column vector of 
%     quantiles X at the cumulative probabilities P. This function automatically 
%     identifies extreme observations and evaluates the appropriate Pareto tail
%     function. A simple linear interpolation is performed for probabilities in 
%     P that fall between the lower and upper Pareto tails. 
%
%     GETICDF does not estimate an inverse CDF, but rather applies an existing
%     quantile function to a cumulative probability vector P, making it 
%     well-suited to copula and Monte Carlo simulation methods.
%

   if length(P) ~= numel(P)
      error('Input probability P must be a vector.');
   else
      if size(P,1) == 1  % Is it a row vector?
         P = P(:);       % Then convert it to a column vector.
      end
   end

   xData = nan(numel(P),1);
%
%  Evaluate the Pareto tail functions (if any).
%
   if (pLower > 0) && ~isLowerTailEmpty
      isInLowerTail        = P <= pLower;  % Lower tail observations.
      probabilities        = 1 - P(isInLowerTail)/pLower;
      xData(isInLowerTail) = QL - gpinv(probabilities, lowerTail, lowerSigma);
   end

   if (pUpper < 1) && ~isUpperTailEmpty
      isInUpperTail        = P >= pUpper;  % Upper tail observations.
      probabilities        = (P(isInUpperTail) - pUpper)/(1 - pUpper);
      xData(isInUpperTail) = QU + gpinv(probabilities, upperTail, upperSigma);
   end
%
%  Interpolate into the center of the distribution with the sample CDF.
%
   iCDF       = ( ((0.9*pLower) <= F) & (F <= (1 - (0.9*(1 - pUpper)))) );
   isInCenter = ~(isInUpperTail | isInLowerTail);

   xData(isInCenter) = interp1(F(iCDF), X(iCDF), P(isInCenter), 'linear', NaN);
%
%  Perform some final NaN and out-of-bounds cleanup. This is usually necessary 
%  only in the event that no Pareto tails have been specified (i.e., no 
%  extrapolation mechanism exists).
%
   iNaNs = isnan(xData);

   xData((P <= 0) & iNaNs) = xMin;
   xData((P >= 1) & iNaNs) = xMax;

   xData(P < 0) = xMin;
   xData(P > 1) = xMax;

   end    % Nested "getICDF" function.

%
%  Identify the "out-of-sample" CDF evaluation function.
%
   function P = getCDF(xData)
%  GETCDF Evaluate and return probabilities of the CDF.
%     P = GETCDF(X) evaluates the cumulative distribution function (CDF) 
%     estimated above, returning a column vector of cumulative probabilities P 
%     at the observations in the data vector X. This function automatically 
%     identifies extreme observations and evaluates the appropriate Pareto tail
%     function. A simple linear interpolation is performed for observations of 
%     X that fall between the lower and upper Pareto tails. 
%
%     GETCDF does not estimate a CDF, but rather applies an estimated CDF to a
%     data vector X, making it well-suited to copula and Monte Carlo simulation
%     methods. Also, the Pareto tails are evaluated by invoking functions, and
%     allow extrapolation beyond the original historical record.
%

   if length(xData) ~= numel(xData)
      error('Input data X must be a vector.');
   else
      if size(xData,1) == 1  % Is it a row vector?
         xData = xData(:);   % Then convert it to a column vector.
      end
   end

   P = nan(numel(xData),1);
%
%  Apply the smoothing function to the data vector X. Although all elements of
%  X are accounted for, any observations of X that fall in the Pareto tails 
%  will be over-written.
%
   P = interp1(X, F, xData, 'linear', NaN);
%
%  Evaluate the Pareto tail functions (if any).
%
   if (pLower > 0) && ~isLowerTailEmpty
      isInLowerTail    = xData <= QL;  % Lower tail observations.
      probabilities    = gpcdf(QL - xData(isInLowerTail), lowerTail, lowerSigma);
      P(isInLowerTail) = pLower * (1 - probabilities);
   end

   if (pUpper < 1) && ~isUpperTailEmpty
      isInUpperTail    = xData >= QU;  % Upper tail observations.
      probabilities    = gpcdf(xData(isInUpperTail) - QU, upperTail, upperSigma);
      P(isInUpperTail) = pUpper + (1 - pUpper) * probabilities;
   end
%
%  Perform some final NaN and out-of-bounds cleanup. This is usually necessary 
%  only in the event that no Pareto tails have been specified (i.e., no 
%  extrapolation mechanism exists).
%
   iNaNs = isnan(P);

   P((xData < xMin) & iNaNs) = 0;
   P((xData > xMax) & iNaNs) = 1;

   P(P < 0) = 0;
   P(P > 1) = 1;

   end    % Nested "getCDF" function.

%
%  Identify the default "in-sample" smoothed empirical CDF function.
%
   function [F,X] = smoothECDF(x)
%  SMOOTHCDF Smooth the empirical CDF by linear interpolation.
%     [F,XI] = SMOOTHCDF(X) calculates the empirical cumulative distribution 
%     function (CDF) from the observed (historical, or in-sample) input data 
%     vector X. The cumulative probabilities F of the empirical CDF are then 
%     smoothed by linear interpolation at the unique input data observations XI.
%
%     To ensure that the smoothed CDF covers the closed interval [0,1], the 
%     first and last elements of the output quantile vector XI are linearly 
%     extrapolated using the adjacent slope of the piecewise linear CDF.
%

   [F, X]   = ecdf(x);                % Empirical CDF.
   F(2:end) = F(2:end) - diff(F)/2;   % Interpolate at each interior CDF step.

%
%  Perform the extrapolation. To ensure the probability = 0 for any observation
%  x <= X(1) = min(X), the extrapolation at low probabilities uses the slope of 
%  the linear segment immediately to the right. Since the first two elements of 
%  X returned by ECDF are the same, over-write the first element X(1) with this
%  extrapolated value. 
%
%  To ensure the probability = 1 for any observation x >= X(end) = max(X), the 
%  extrapolation at high probabilities uses the slope of the linear segment 
%  immediately to the left. In this segment of the probability curve, we append 
%  the extrapolated quantile and the corresponding probability = 1.
%

   n    = length(X);
   X(1) = X(2) - F(2)*(X(3) - X(2))/(F(3) - F(2));
   Xn   = X(n) + (1 - F(n))*(X(n) - X(n-1))/(F(n) - F(n-1));
   X = [X ; Xn];
   F = [F ;  1];

   end

end       % Primary "cdfByGPD" function.
