function [DoF, Rho, nLLF] = tCopulaFit(u)
% Estimate the scalar degrees-of-freedom (DoF) parameter and the linear 
% correlation matrix (Rho) of the t-copula.
%
% Notice that there are several ways of doing this:
%
% (1) We could minimize the full negative log-likelihood directly, but that 
%     tends to be slow because of all the calls to inverse of a t-distribution 
%     (see function TINV). Also, the optimizer can have trouble because there
%     is often a long flat valley, and in high dimensions that's hard to deal 
%     with. Thus, it's better to maximize the likelihood with a two-step 
%     iteration.
%
% (2) We could use Bouye's et al's iterative solution for Rho derived by
%     differentiating the negative log-likelihood function with respect to the 
%     correlation matrix, given a fixed degree-of-freedom DoF. 
%
%     This too seems to be motivated as a profile likelihood, but in fact it 
%     is not, because it does not converge to the conditional MLE for Rho given
%     DoF. However, the estimator may be valid in it's own right, although
%     it is not the MLE.
%
% (3) We could use the approach recommended by Zeevi and Mashal, which estimates
%     the linear correlation matrix Rho by applying a robust sine transform to a
%     dependence measure such as Kendall's or Spearman's sample rank correlation
%     matrix. Given this estimate of Rho, we then minimize the negative 
%     log-likelihood function with respect to the DoF parameter alone.
%
%     Although this approach is not really a profile likelihood, it too seems to
%     be motivated by such an approach, and may also be a valid method in its
%     own right. Since this method does not require matrix inversion, it has the
%     advantage of being very robust and numerically stable in the presence of
%     close-to-singular matrices.
%

%   Copyright 1993-2005 The MathWorks, Inc.
%   $Revision$   $Date$

% 
% With data that are reasonably fit by a small DoF, the estimate tends to have 
% decent precision. The bad news is that with data that are reasonably fit with
% a somewhat larger DoF (25, say), the precision of the estimate is very poor. 
% But the good news is that it doesn't matter much because the correlation 
% estimate is almost independent of the DoF estimate in these cases, and the 
% distribution doesn't change all that much with respect to DoF once it is 
% larger than, say, 100 -- you might as well fit a normal.
%

% Setting this upper limit saves some iterations in the estimation. If the
% estimate ends up right at this upper limit, increase it to, say, 1000.

DoF_lower =   2;
DoF_upper =  30;

%
% Use FMINBND to maximize the profile likelihood for DoF.
%
options1 = optimset('TolX'      , 1e-2, 'Display', 'off');
options2 = optimset('TolX'      , 1e-3, 'TolFun' , 1e-3, 'MaxIter', 1000, ...
                    'LargeScale','off', 'Display', 'off');

[DoF, nLLF] = fminbnd(@profileNegloglike, DoF_lower, DoF_upper, options1);
[xxx, Rho]  = profileNegloglike(DoF);

   function [nLLF, Rho] = profileNegloglike(nu)
%  Profile negative log-likelihood for the degree-of-freedom (DoF).

%  Transform uniformly-distributed variates to t-distributed variates
%  and compute initial estimate for the linear correlation matrix (Rho).

   t   = tinv(u, nu);
   Rho = corr(t);

%  Compute Rho using conditional maximum likelihood, given nu.

   objFun         = @(params) negloglike(nu, tomatrix(params), t);
   [params, nLLF] = fminunc(objFun, tovector(chol(Rho)), options2);

   R   = tomatrix(params);
   Rho = R'*R;
   end

end    % End of primary function tCopulaFit.


function nLLF = negloglike(nu, R, t)
% Compute negative log-likelihood at nu and R = chol(Rho)

[n,d] = size(t);
R     = R ./ repmat(sqrt(sum(R.^2,1)),d,1);
tRinv = t / R;
nLLF  = - n * (gammaln((nu + d)/2) - gammaln(nu/2)) ...
        + n * d *(gammaln((nu + 1)/2) - gammaln(nu/2)) ...
        + n * sum(log(abs(diag(R)))) ...
        + ((nu + d)/2)*sum(log(1 + sum(tRinv.^2, 2)/nu)) ...
        - ((nu + 1)/2)*sum(sum(log(1 + (t.^2)/nu),2),1);
end


function b = tovector(A)
% Convert the cholesky factor of a correlation matrix to upper triangle vector
% form.  Columns of the cholesky factor have unit norm, so they reduce to
% direction angles, consisting of one fewer element.
m = size(A,1);
Angles = zeros(m);
for i = 1:m-1
    Angles(i,:) = atan2(A(i,:),A(i+1,:));
    A(i+1,:) = A(i+1,:) ./ cos(Angles(i,:));
end
b = Angles(triu(true(m),1));
end


function A = tomatrix(b)
% Convert the cholesky factor of a correlation matrix from upper triangle vector
% form.  Columns of the cholesky factor have unit norm, so the columns can be
% recreated from direction angles.
m = (1 + sqrt(1+8*length(b)))/2;
Cosines = zeros(m);
Cosines(triu(true(m),1)) = cos(b);
Sines = ones(m);
Sines(triu(true(m),1)) = sin(b);
A = [ones(1,m); Cosines(1:m-1,:)] .* flipud(cumprod(flipud(Sines)));
end
